#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 22 12:05:31 2023

@author: weiss
"""
#https://www.math.univ-toulouse.fr/~weiss/Temp/cours1.py


#%% Importing libraries
import numpy as np
from skimage import data, io, filters, util
import matplotlib.pyplot as plt

#%% Loading a few images
#plant = skimage.io.imread("")
plant = io.imread("/home/weiss/Bureau/COURS/cours_m2bio_image/codes/membrane.png")
cell = data.cell()
retina = data.retina()  #https://scikit-image.org/docs/stable/auto_examples/data/plot_scientific.html#sphx-glr-auto-examples-data-plot-scientific-py

plt.imshow(plant)
plt.title("A weird title")
plt.show()
plt.imshow(cell)
plt.title("Cell image")
plt.show()
plt.imshow(retina)
plt.title("Retina image")
plt.show()

#%% Exploring the type of the data
print(plant.shape)
print(plant.dtype) # Here, the data is just integers

print(cell.shape)
print(retina.shape)

#%% Rescaling in [0,1]
plant = np.float32(plant/255.) # Changing the type of the data to allow for more complex operations
cell = (cell-cell.min())/(cell.max()-cell.min())

#%% Channel selection, Cropping, flipping, downsizing
plt.imshow(plant[:,:,1]) #Keeping the green channel only (RGB)
plt.title("Green channel of plant (default colormap)")
plt.show()

plt.imshow(plant[:,:,1],cmap="gray") #Keeping the green channel only (RGB), with a gray colormap instead of the default
plt.title("Green channel of plant (gray colormap)")
plt.show()

plt.imshow(plant[:,:,1],cmap="gray",vmin = 0, vmax=0.5) #Keeping the green channel only (RGB), specifying the range of the lookup table
plt.title("Green channel of plant (fixing the range)")
plt.show()

print(plant[:,:,1].shape) 
plt.imshow(plant[100:200,100:150,:]) #cropping a rectangular part of the image
plt.title("Crop of plant")
plt.show()
print(plant[100:200,100:150,:].shape)

plt.imshow(plant[100:200:4,::4,:]) # Downsizing by taking 1 pixel out of 16
plt.title("Downsampling by 4")
plt.show()

plt.imshow(plant[::-1,::1,:]) # Flipping the y-axis
plt.title("Reverse order in y direction")
plt.show()

plt.imshow(np.swapaxes(plant,0,1)) # rotating by 90 degrees
plt.title("Swapping axes (90 ° rotation)")
plt.show()

#%% Image representation 
#%matplotlib inline
#%matplotlib auto #for interactive plots
fig = plt.figure()  #displaying a simple image as a surface
ax = plt.axes(projection ='3d')
[X,Y] = np.meshgrid(np.arange(cell.shape[1]),np.arange(cell.shape[0]))
ax.plot_surface(X,Y,cell)
plt.title("Surface plot of cell")
plt.show()

cell_smoothed = filters.gaussian(cell,10) # Zg = a Gaussian filtering of Z (10 means that we average over 10 pixels)
fig = plt.figure()   #displaying a filtered image as a surface
ax = plt.axes(projection ='3d')
[X,Y] = np.meshgrid(np.arange(cell.shape[1]),np.arange(cell.shape[0]))
ax.plot_surface(X,Y,cell_smoothed)
plt.title("Surface plot of smoothed cell")
plt.show()

ctr = plt.contour(X,Y,cell,[0.2,0.4,0.6,0.8]) # Displaying the level lines of the image (as an IGN map)
plt.clabel(ctr, inline=1, fontsize=10)
plt.title("Contour plot of cell (level lines)")
plt.show()

#%% Pointwise operations
#Inverting the contrasts
plt.imshow(cell,cmap="gray")
plt.title("Cell")
plt.show()

plt.imshow(1-cell,cmap="gray") 
plt.title("-Cell (reverse contrasts)")
plt.show()

# Thresholding
threshold = 0.3
plt.imshow(cell > threshold,cmap="gray")
plt.title("Threshold: %1.2e"%threshold)
plt.show()

# Contrast change
plt.imshow(cell**2,cmap="gray")
plt.title("Contrast change (square)")
plt.show()

plt.imshow(np.sqrt(cell),cmap="gray")
plt.title("Contrast change (square root)")
plt.show()

plt.imshow(np.sin(6*2*np.pi*cell),cmap="gray")
plt.title("Contrast change (sinusoid)")
plt.show()

# Quantization
n_levels = 4
cell_quantized = np.floor(cell*n_levels)
plt.imshow(cell_quantized,cmap="gray")
plt.title("Quantized on %i levels"%n_levels)
plt.show()


#%% Convolutions (possibly the hardest but most important part)

#Adding noise to an image to simulate a degradation
noise = np.random.randn(cell.shape[0],cell.shape[1])
cell_noisy = cell + noise
plt.imshow(cell_noisy)
plt.title("A very noisy cell")
plt.show()

plt.imshow(cell_noisy, cmap="gray",vmin=0, vmax = 1)
plt.title("A very noisy cell (with better visualization)")
plt.show()

#Image smoothing (Gaussian filtering)
cell_smoothed = filters.gaussian(cell_noisy,5) # Zg = a Gaussian filtering of Z (10 means that we average over 10 pixels)
plt.imshow(cell_smoothed, cmap="gray",vmin=0, vmax = 1)
plt.title("Gaussian filtered noisy cell")
plt.show()

# Defining a convolution function
def convolution(image,filter):
    n1 = image.shape[0]
    n2 = image.shape[1]
    
    m1 = filter.shape[0]
    m2 = filter.shape[1]
    
    if m1%2 == 0:
        print("Size should be odd")
        return 0
    
    result = np.zeros_like(image)
    for x1 in range(n1):
        for x2 in range(n2):
            for y1 in range(m1):
                for y2 in range(m2):
                    c1 = x1 + y1 - m1//2
                    c2 = x2 + y2 - m2//2
                    
                    if c1>=0 and c1<n1 and c2>=0 and c2<n2:
                        result[x1,x2] = result[x1,x2] + image[c1,c2]*filter[y1,y2]
    return result

#Edge detection x-direction
filtery = np.zeros((3,3))
filtery[1,1] = -1
filtery[2,1] = 1

celly = convolution(cell,filtery)
plt.imshow(celly, cmap="gray")
plt.title("y gradient of cell")
plt.show()

filterx = np.zeros((3,3))
filterx[1,1] = -1
filterx[1,2] = 1

cellx = convolution(cell,filterx)
plt.imshow(cellx, cmap="gray")
plt.title("x gradient of cell")
plt.show()

plt.quiver(cellx[::4,::4],celly[::4,::4])
plt.title("Gradient field")
plt.show()

norm_gradient = np.sqrt(cellx**2 + celly**2)
plt.imshow(norm_gradient)
plt.title("Gradient amplitude")
plt.show()

plt.imshow(norm_gradient > 0.045)
plt.title("Boundaries")
plt.show()

#Z_average = filter.average()


#%% Morphological operations
cell_sp = util.random_noise(cell, mode = "s&p", amount = 0.6) # Adding salt and pepper noise
plt.imshow(cell_sp, cmap = "gray")
plt.title("salt & pepper noise")
plt.show()

#Median filtering
from skimage.morphology import disk
cell_denoised = filters.median(cell_sp,disk(7))
plt.imshow(cell_denoised, cmap = "gray")
plt.title("Median filtered image")
plt.show()

# Binary morphology (to clean masks)
from skimage.morphology import (erosion, dilation, opening, closing)
mask = (cell>=0.3) # A naive thresholding with a bad threshold
plt.imshow(mask, cmap = "gray")
plt.title("Mask")
plt.show()

#Erosion
strel = disk(6) #structuring element
print(strel)
eroded = erosion(mask, strel)
plt.imshow(eroded, cmap = "gray")
plt.title("Eroded mask")
plt.show()

#Dilation
strel = disk(6) #structuring element
print(strel)
dilated = dilation(mask, strel)
plt.imshow(dilated, cmap = "gray")
plt.title("Dilated mask")
plt.show()

#Opening
strel = disk(6) #structuring element
print(strel)
opened = opening(mask, strel)
plt.imshow(opened, cmap = "gray")
plt.title("Opened = (eroded + dilated) mask")
plt.show()

#Closing
strel = disk(6) #structuring element
print(strel)
closed = closing(mask, strel)
plt.imshow(closed, cmap = "gray")
plt.title("Closed = (dilated + eroded) mask")
plt.show()